/**
 * Gainspan wi-fi module library for LPC1114 (LPCXpresso)
 * code from http://mbed.org/users/gsfan/code/GSwifi/
 * Modified for Arduino, 2012 Galileo-7, S.
 */

/*
 * Gainspan wi-fi module library for mbed
 * Copyright (c) 2012 gsfan
 * Released under the MIT License: http://mbed.org/license/mit
 */

/** @file
 * @brief Gainspan wi-fi module library for mbed
 * GS1011MIC, GS1011MIP, GainSpan WiFi Breakout, etc.
 * module configuration: ATB=115200
 */

#include "dbg.h"
#include "LPC11xx.h"
#include "GSwifi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define _gs_getc() LPC_UART->RBR
#define _gs_putc(c) while(!(LPC_UART->LSR & (1<<5))); LPC_UART->THR = c

uint32_t millis();
void wait_ms(int ms);

int _rts;
volatile int _connect;
volatile enum GSSTATUS _status;
volatile int _gs_ok, _gs_failure, _gs_enter;
enum GSMODE _gs_mode;
int _escape;
int _cid, _rssi;
IpAddr _ipaddr, _netmask, _gateway, _nameserver, _resolv;
struct Host _from, _to;
char _mac[6];
struct RingBuffer *_buf_cmd;
struct GS_Socket _gs_sock[16];
char _year, _month, _day, _hour, _min, _sec;


volatile unsigned long timer_time = 0;
void Timer_start () {
	timer_time = millis();
}
void Timer_stop () {
}
void Timer_reset () {
	Timer_start();
}
unsigned long Timer_read_ms () {
	return millis() - timer_time;
}

void init_uart (int baud, int flow) {
	int dl;
	LPC_IOCON->PIO1_6 |= 0x01; // RXD
	LPC_IOCON->PIO1_7 |= 0x01; // TXD
	LPC_SYSCON->SYSAHBCLKCTRL |= (1<<12);
	LPC_SYSCON->UARTCLKDIV = 0x01;
	dl = (SystemCoreClock * LPC_SYSCON->SYSAHBCLKDIV) / (16 * baud * LPC_SYSCON->UARTCLKDIV);
	LPC_UART->LCR |= (1<<7); // DLAB
	LPC_UART->DLM = dl / 256;
	LPC_UART->DLL = dl % 256;
	LPC_UART->LCR &= ~(1<<7);
	LPC_UART->LCR = (3<<0); // 8bit
	LPC_UART->FCR = (3<<6)|(1<<2)|(1<<1)|(1<<0); // fifo 14, clear, fifo on
	if (flow) {
		LPC_IOCON->PIO0_7 |= 0x01; // CTS
		LPC_IOCON->PIO1_5 |= 0x01; // RTS
		LPC_UART->MCR = (1<<7)|(1<<6); // auto RTS, CTS
	}
	NVIC_EnableIRQ(UART_IRQn);
	LPC_UART->IER = (1<<0); // rx int
}

void _gs_printf (const char *buf) {
    int i;

    for (i = 0; i < strlen(buf); i ++) {
        _gs_putc(buf[i]);
    }
}

void GSwifi (int baud, int flow) {
    _connect = 0;
    _status = GSSTAT_READY;
    _escape = 0;
    _gs_mode = GSMODE_COMMAND;
    _buf_cmd = ring_new(GS_CMD_SIZE);

    init_uart(baud, flow);
    _rts = flow;
}

// uart interrupt
void UART_IRQHandler () {
    static int len, mode;
    static char tmp[20];
    char flg, dat;

    flg = LPC_UART->LSR;
    dat = _gs_getc();

    if (flg & ((1 << 7)|(1 << 3)|(1 << 4))) return;
//    DBG("%02x ", dat);

    switch (_gs_mode) {
    case GSMODE_COMMAND: // command responce
        if (_escape) {
            // esc
            switch (dat) {
            case 'O':
                DBG("ok\r\n");
                _gs_ok = 1;
                break;
            case 'F':
                DBG("failure\r\n");
                _gs_failure = 1;
                break;
            case 'S':
                DBG("GSMODE_DATA_RX\r\n");
                _gs_mode = GSMODE_DATA_RX;
                mode = 0;
                break;
            case 'u':
                DBG("GSMODE_DATA_RXUDP\r\n");
                _gs_mode = GSMODE_DATA_RXUDP;
                mode = 0;
                break;
            case 'Z':
            case 'H':
                DBG("GSMODE_DATA_RX_BULK\r\n");
                _gs_mode = GSMODE_DATA_RX_BULK;
                mode = 0;
                break;
            case 'y':
                DBG("GSMODE_DATA_RXUDP_BULK\r\n");
                _gs_mode = GSMODE_DATA_RXUDP_BULK;
                mode = 0;
                break;
            default:
                DBG("unknown [ESC] %02x\r\n", dat);
                break;
            }
            _escape = 0;
        } else {
            if (dat == 0x1b) {
                _escape = 1;
            } else
            if (dat != '\r') {
                // command
                ring_put(_buf_cmd, dat);
                if (dat == '\n') _gs_enter ++;
            }
        }
        break;

    case GSMODE_DATA_RX:
    case GSMODE_DATA_RXUDP:
        if (mode == 0) {
            // cid
            _cid = x2i(dat);
            _gs_sock[_cid].received = 0;
            mode ++;
            if (_gs_mode == GSMODE_DATA_RX) {
                mode = 3;
            }
            len = 0;
        } else
        if (mode == 1) {
            // ip
            if ((dat < '0' || dat > '9') && dat != '.') {
                int ip1, ip2, ip3, ip4;
                tmp[len] = 0;
                sscanf(tmp, "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4);
                _from.ip = (ip1<<24)|(ip2<<16)|(ip3<<8)|ip4;
                mode ++;
                len = 0;
                break;
            }
            tmp[len] = dat;
            len ++;
        } else
        if (mode == 2) {
            // port
            if (dat < '0' || dat > '9') {
                tmp[len] = 0;
                _from.port = atoi(tmp);
                mode ++;
                len = 0;
                break;
            }
            tmp[len] = dat;
            len ++;
        } else
        if (_escape) {
            // esc
            switch (dat) {
            case 'E':
                DBG("recv ascii %d\r\n", _cid);
                _gs_sock[_cid].received = 1;
                _gs_mode = GSMODE_COMMAND;
                // recv interrupt
                if (_gs_sock[_cid].protocol == GSPROT_HTTPGET && _gs_sock[_cid].onGsReceive != NULL) {
                    _gs_sock[_cid].onGsReceive(_cid, ring_use(_gs_sock[_cid].data));
                    _gs_sock[_cid].received = 0;
                }
                break;
            default:
                DBG("unknown <ESC> %02x\r\n", dat);
                break;
            }
            _escape = 0;
        } else {
            if (dat == 0x1b) {
                _escape = 1;
            } else {
                // data
                ring_put(_gs_sock[_cid].data, dat);
                len ++;
                if (len < GS_DATA_SIZE && ring_available(_gs_sock[_cid].data) == 0) {
                	// buffer full
                	if (_gs_sock[_cid].onGsReceive != NULL) {
                    	_gs_sock[_cid].onGsReceive(_cid, ring_use(_gs_sock[_cid].data));
                    }
                }
            }
        }
        break;

    case GSMODE_DATA_RX_BULK:
    case GSMODE_DATA_RXUDP_BULK:
//        DBG("%c", dat);
        if (mode == 0) {
            // cid
            _cid = x2i(dat);
            _gs_sock[_cid].received = 0;
            mode ++;
            if (_gs_mode == GSMODE_DATA_RX_BULK) {
                mode = 3;
            }
            len = 0;
        } else
        if (mode == 1) {
            // ip
            if ((dat < '0' || dat > '9') && dat != '.') {
                int ip1, ip2, ip3, ip4;
                tmp[len] = 0;
                sscanf(tmp, "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4);
                _from.ip = (ip1<<24)|(ip2<<16)|(ip3<<8)|ip4;
                mode ++;
                len = 0;
                break;
            }
            tmp[len] = dat;
            len ++;
        } else
        if (mode == 2) {
            // port
            if (dat < '0' || dat > '9') {
                tmp[len] = 0;
                _from.port = atoi(tmp);
                mode ++;
                len = 0;
                break;
            }
            tmp[len] = dat;
            len ++;
        } else
        if (mode == 3) {
            // length
            tmp[len] = dat;
            len ++;
            if (len >= 4) {
                tmp[len] = 0;
                len = atoi(tmp);
                mode ++;
                break;
            }
        } else
        if (mode == 4) {
            // data
            if (_gs_sock[_cid].data != NULL) {
                ring_put(_gs_sock[_cid].data, dat);
            }
            len  --;
            if (len && ring_available(_gs_sock[_cid].data) == 0) {
            	// buffer full
            	if (_gs_sock[_cid].onGsReceive != NULL) {
                	_gs_sock[_cid].onGsReceive(_cid, ring_use(_gs_sock[_cid].data));
                }
            }
            if (len == 0) {
                DBG("recv binary %d\r\n", _cid);
                _gs_sock[_cid].received = 1;
                _escape = 0;
                _gs_mode = GSMODE_COMMAND;
                // recv interrupt
                if (_gs_sock[_cid].protocol == GSPROT_HTTPGET && _gs_sock[_cid].onGsReceive != NULL) {
                    _gs_sock[_cid].onGsReceive(_cid, ring_use(_gs_sock[_cid].data));
                    _gs_sock[_cid].received = 0;
                }
            }
        }
        break;

    }
}

int command (const char *cmd, enum GSRESPONCE res, int timeout) {
    int i;

    if (! cmd) {
        // dummy CR+LF
        _gs_printf("\r\n");
        for (i = 0; i < 10; i ++) {
        	wait_ms(10);
        	poll();
        	ring_clear(_buf_cmd);
        }
        return 0;
    }

    ring_clear(_buf_cmd);
    _gs_ok = 0;
    _gs_failure = 0;
    _gs_enter = 0;
    for (i = 0; i < strlen(cmd); i ++) {
        _gs_putc(cmd[i]);
    }
    _gs_putc('\r');
    _gs_putc('\n');
    DBG("command: %s\r\n", cmd);
    if (strlen(cmd) == 0) return 0;
    wait_ms(50);
    if (timeout) {
        return cmdResponse(res, timeout);
    } else {
        return 0;
    }
}

int cmdResponse (enum GSRESPONCE res, int ms) {
    int i, n = 0, flg = 0;
    char buf[GS_CMD_SIZE];

    Timer_start();
    for (;;) {
        // recv response
        i = 0;
        while (i < sizeof(buf)) {
            if (ring_use(_buf_cmd)) {
                ring_get(_buf_cmd, &buf[i]);
                if (buf[i] == '\n') {
                    break;
                }
                i ++;
            }
            if (Timer_read_ms() > ms) {
            	Timer_stop();
                DBG("timeout\r\n");
                return -1;
            }
        }
        _gs_enter = 0;
        if (i == 0) continue;
        buf[i] = 0;
        DBG("response: %s\r\n", buf);
        Timer_stop();

        if (strcmp(buf, "OK") == 0) {
            _gs_ok = 1;
        } else
        if (strncmp(buf, "ERROR", 5) == 0) {
            _gs_failure = 1;
        }

        switch(res) {
        case GSRES_NORMAL:
            flg = 1;
            break;
        case GSRES_WPS:
            if (n == 0 && strncmp(buf, "SSID", 4) == 0) {
                n ++;
            } else
            if (n == 1 && strncmp(buf, "CHANNEL", 7) == 0) {
                n ++;
            } else
            if (n == 2 && strncmp(buf, "PASSPHRASE", 10) == 0) {
                n ++;
                flg = 1;
            }
            break;
        case GSRES_CONNECT:
            if (strncmp(buf, "CONNECT", 7) == 0) {
                _cid = x2i(buf[8]);
                flg = 1;
            }
            break;
        case GSRES_DHCP:
            if (n == 0 && strstr(buf, "SubNet") && strstr(buf, "Gateway")) {
                n ++;
            } else
            if (n == 1) {
                int ip1, ip2, ip3, ip4;
                char *tmp = buf + 1;
                sscanf(tmp, "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4);
                _ipaddr = (ip1<<24)|(ip2<<16)|(ip3<<8)|ip4;
                tmp = strstr(tmp, ":") + 2;
                sscanf(tmp, "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4);
                _netmask = (ip1<<24)|(ip2<<16)|(ip3<<8)|ip4;
                tmp = strstr(tmp, ":") + 2;
                sscanf(tmp, "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4);
                _gateway = (ip1<<24)|(ip2<<16)|(ip3<<8)|ip4;
                n ++;
                flg = 1;
            }
            break;
        case GSRES_MACADDRESS:
            if (buf[2] == ':' && buf[5] == ':') {
                int mac1, mac2, mac3, mac4, mac5, mac6;
                sscanf(buf, "%x:%x:%x:%x:%x:%x", &mac1, &mac2, &mac3, &mac4, &mac5, &mac6);
                _mac[0] = mac1;
                _mac[1] = mac2;
                _mac[2] = mac3;
                _mac[3] = mac4;
                _mac[4] = mac5;
                _mac[5] = mac6;
                flg = 1;
            }
            break;
        case GSRES_DNSLOOKUP:
            if (strncmp(buf, "IP:", 3) == 0) {
                int ip1, ip2, ip3, ip4;
                sscanf(&buf[3], "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4);
                _resolv = (ip1<<24)|(ip2<<16)|(ip3<<8)|ip4;
                flg = 1;
            }
            break;
        case GSRES_HTTP:
            if (buf[0] >= '0' && buf[0] <= 'F') {
                _cid = x2i(buf[0]);
                flg = 1;
            }
            break;
        case GSRES_RSSI:
            if (buf[0] == '-' || (buf[0] >= '0' && buf[0] <= '9')) {
                _rssi = atoi(buf);
                flg = 1;
            }
            break;
        case GSRES_TIME:
            if (buf[0] >= '0' && buf[0] <= '9') {
                int year, month, day, hour, min, sec;
                sscanf(buf, "%d/%d/%d,%d:%d:%d", &day, &month, &year, &hour, &min, &sec);
                _sec = sec;
                _min = min;
                _hour = hour;
                _day = day;
                _month = month - 1;
                _year = year - 1900;
                flg = 1;
            }
            break;
        }

        if ((flg && _gs_ok) || _gs_failure) break;
        Timer_reset();
        Timer_start();
    }

    if (_gs_failure || flg == 0) {
        return -1;
    } else {
        return 0;
    }
}

int connect (enum GSSECURITY sec, char *ssid, char *pass, int dhcp) {
    int r;
    char cmd[GS_CMD_SIZE];

    if (_connect || _status != GSSTAT_READY) return -1;

    command(NULL, GSRES_NORMAL, GS_TIMEOUT);
    if (command("ATE0", GSRES_NORMAL, GS_TIMEOUT)) return -1;
    if (_rts) {
        command("AT&K0", GSRES_NORMAL, GS_TIMEOUT);
        command("AT&R1", GSRES_NORMAL, GS_TIMEOUT);
    }
    command("AT+NMAC=?", GSRES_MACADDRESS, GS_TIMEOUT);
#ifdef GS_BULK
    command("AT+BDATA=1", GSRES_NORMAL, GS_TIMEOUT);
#endif

    disconnect();
    command("AT+WM=0", GSRES_NORMAL, GS_TIMEOUT); // infrastructure
    wait_ms(100);
    if (dhcp) {
        command("AT+NDHCP=1", GSRES_NORMAL, GS_TIMEOUT);
    } else {
        command("AT+NDHCP=0", GSRES_NORMAL, GS_TIMEOUT);
    }

    switch (sec) {
    case GSSEC_NONE:
        command("AT+WAUTH=0", GSRES_NORMAL, GS_TIMEOUT);
        sprintf(cmd, "AT+WA=%s", ssid);
        r = command(cmd, GSRES_NORMAL, GS_TIMEOUT2);
        break;
    case GSSEC_OPEN:
    case GSSEC_WEP:
        sprintf(cmd, "AT+WAUTH=%d", sec);
        command(cmd, GSRES_NORMAL, GS_TIMEOUT);
        sprintf(cmd, "AT+WWEP1=%s", pass);
        command(cmd, GSRES_NORMAL, GS_TIMEOUT);
        wait_ms(100);
        sprintf(cmd, "AT+WA=%s", ssid);
        r = command(cmd, GSRES_DHCP, GS_TIMEOUT2);
        break;
    case GSSEC_WPA_PSK:
    case GSSEC_WPA2_PSK:
        command("AT+WAUTH=0", GSRES_NORMAL, GS_TIMEOUT);
//        sprintf(cmd, "AT+WWPA=%s", pass);
//        command(cmd, GSRES_NORMAL);
        sprintf(cmd, "AT+WPAPSK=%s,%s", ssid, pass);
        command(cmd, GSRES_NORMAL, GS_TIMEOUT2);
        wait_ms(100);
        sprintf(cmd, "AT+WA=%s", ssid);
        r = command(cmd, GSRES_DHCP, GS_TIMEOUT2);
        break;
    case GSSEC_WPS_BUTTON:
        command("AT+WAUTH=0", GSRES_NORMAL, GS_TIMEOUT);
        r = command("AT+WWPS=1", GSRES_WPS, GS_TIMEOUT2);
        if (r) break;
        if (dhcp) {
            r = command("AT+NDHCP=1", GSRES_DHCP, GS_TIMEOUT2);
        }
        break;
    default:
        DBG("Can't use security\r\n");
        r = -1;
        break;
    }

    if (r == 0) _connect = 1;
    return r;
}

int adhock (enum GSSECURITY sec, char *ssid, char *pass, IpAddr ipaddr, IpAddr netmask) {
    int r;
    char cmd[GS_CMD_SIZE];

    if (_connect || _status != GSSTAT_READY) return -1;

    command(NULL, GSRES_NORMAL, GS_TIMEOUT);
    if (command("ATE0", GSRES_NORMAL, GS_TIMEOUT)) return -1;
    if (_rts) {
        command("AT&K0", GSRES_NORMAL, GS_TIMEOUT);
        command("AT&R1", GSRES_NORMAL, GS_TIMEOUT);
    }
    disconnect();
    command("AT+NMAC=?", GSRES_MACADDRESS, GS_TIMEOUT);
#ifdef GS_BULK
    command("AT+BDATA=1", GSRES_NORMAL, GS_TIMEOUT);
#endif

    command("AT+WM=1", GSRES_NORMAL, GS_TIMEOUT); // adhock
    wait_ms(100);
    command("AT+NDHCP=0", GSRES_NORMAL, GS_TIMEOUT);
    setAddress(ipaddr, netmask, ipaddr, ipaddr);

    switch (sec) {
    case GSSEC_NONE:
        command("AT+WAUTH=0", GSRES_NORMAL, GS_TIMEOUT);
        sprintf(cmd, "AT+WA=%s", ssid);
        r = command(cmd, GSRES_NORMAL, GS_TIMEOUT2);
        break;
    case GSSEC_OPEN:
    case GSSEC_WEP:
        sprintf(cmd, "AT+WAUTH=%d", sec);
        command(cmd, GSRES_NORMAL, GS_TIMEOUT);
        sprintf(cmd, "AT+WWEP1=%s", pass);
        command(cmd, GSRES_NORMAL, GS_TIMEOUT);
        wait_ms(100);
        sprintf(cmd, "AT+WA=%s", ssid);
        r = command(cmd, GSRES_NORMAL, GS_TIMEOUT2);
        break;
    default:
        DBG("Can't use security\r\n");
        r = -1;
        break;
    }

    if (r == 0) _connect = 1;
    return r;
}

int limitedap (enum GSSECURITY sec, char *ssid, char *pass, IpAddr ipaddr, IpAddr netmask) {
    int r;
    char cmd[GS_CMD_SIZE];

    if (_connect || _status != GSSTAT_READY) return -1;

    command(NULL, GSRES_NORMAL, GS_TIMEOUT);
    if (command("ATE0", GSRES_NORMAL, GS_TIMEOUT)) return -1;
    if (_rts) {
        command("AT&K0", GSRES_NORMAL, GS_TIMEOUT);
        command("AT&R1", GSRES_NORMAL, GS_TIMEOUT);
    }
    disconnect();
    command("AT+NMAC=?", GSRES_MACADDRESS, GS_TIMEOUT);
#ifdef GS_BULK
    command("AT+BDATA=1", GSRES_NORMAL, GS_TIMEOUT);
#endif

    command("AT+WM=2", GSRES_NORMAL, GS_TIMEOUT); // limited ap
    wait_ms(1000);
    command("AT+NDHCP=0", GSRES_NORMAL, GS_TIMEOUT);
    setAddress(ipaddr, netmask, ipaddr, ipaddr);
    if (command("AT+DHCPSRVR=1", GSRES_NORMAL, GS_TIMEOUT)) return -1;
    if (command("AT+DNS=1,setup", GSRES_NORMAL, GS_TIMEOUT)) return -1;

    switch (sec) {
    case GSSEC_NONE:
        command("AT+WAUTH=0", GSRES_NORMAL, GS_TIMEOUT);
        sprintf(cmd, "AT+WA=%s", ssid);
        r = command(cmd, GSRES_NORMAL, GS_TIMEOUT2);
        break;
    case GSSEC_OPEN:
    case GSSEC_WEP:
        sprintf(cmd, "AT+WAUTH=%d", sec);
        command(cmd, GSRES_NORMAL, GS_TIMEOUT);
        sprintf(cmd, "AT+WWEP1=%s", pass);
        command(cmd, GSRES_NORMAL, GS_TIMEOUT);
        wait_ms(100);
        sprintf(cmd, "AT+WA=%s", ssid);
        r = command(cmd, GSRES_NORMAL, GS_TIMEOUT2);
        break;
    default:
        DBG("Can't use security\r\n");
        r = -1;
        break;
    }

    if (r == 0) _connect = 1;
    return r;
}

int disconnect () {
    int i;

    _connect = 0;
    for (i = 0; i < 16; i ++) {
        _gs_sock[i].connect = 0;
    }
    command("AT+NCLOSEALL", GSRES_NORMAL, GS_TIMEOUT);
    command("AT+WD", GSRES_NORMAL, GS_TIMEOUT);
    command("AT+NDHCP=0", GSRES_NORMAL, GS_TIMEOUT);
    wait_ms(100);
    return 0;
}

int setAddress (IpAddr ipaddr, IpAddr netmask, IpAddr gateway, IpAddr nameserver) {
    int r;
    char cmd[GS_CMD_SIZE];
    uint8_t *ip = (uint8_t *)&ipaddr, *nm = (uint8_t *)&netmask, *gw = (uint8_t *)&gateway, *ns = (uint8_t *)&nameserver;

    if (ipaddr == 0) {
        if (command("AT+NDHCP=1", GSRES_DHCP, GS_TIMEOUT2)) return -1;
        if (_ipaddr == 0) return -1;
        return 0;
    }

    command("AT+NDHCP=0", GSRES_NORMAL, GS_TIMEOUT);
    wait_ms(100);

    sprintf(cmd, "AT+NSET=%d.%d.%d.%d,%d.%d.%d.%d,%d.%d.%d.%d",
        ip[3], ip[2], ip[1], ip[0],
        nm[3], nm[2], nm[1], nm[0],
        gw[3], gw[2], gw[1], gw[0]);
    r = command(cmd, GSRES_NORMAL, GS_TIMEOUT);
    if (r) return -1;
    _ipaddr = ipaddr;
    _netmask = netmask;
    _gateway = gateway;

    if (ipaddr != nameserver) {
        sprintf(cmd, "AT+DNSSET=%d.%d.%d.%d",
            ns[3], ns[2], ns[1], ns[0]);
        r = command(cmd, GSRES_NORMAL, GS_TIMEOUT);
    }
    return r;
}

int getAddress (IpAddr *ipaddr, IpAddr *netmask, IpAddr *gateway, IpAddr *nameserver) {
    *ipaddr = _ipaddr;
    *netmask = _netmask;
    *gateway = _gateway;
    *nameserver = _nameserver;
    return 0;
}

int getHostByName (const char* name, IpAddr *addr) {
    char cmd[GS_CMD_SIZE];

    if (! _connect || _status != GSSTAT_READY) return -1;

    sprintf(cmd, "AT+DNSLOOKUP=%s", name);
    if (command(cmd, GSRES_DNSLOOKUP, GS_TIMEOUT)) return -1;

    *addr = _resolv;
    return 0;
}

int getHostByName2 (struct Host *host) {
    char cmd[GS_CMD_SIZE];

    if (! _connect || _status != GSSTAT_READY) return -1;

    sprintf(cmd, "AT+DNSLOOKUP=%s", host->name);
    if (command(cmd, GSRES_DNSLOOKUP, GS_TIMEOUT)) return -1;

    host->ip = _resolv;
    return 0;
}

int setRFPower (int power) {
    char cmd[GS_CMD_SIZE];

    if (power < 0 || power > 7) return -1;

    sprintf(cmd, "AT+WP=%d", power);
    return command(cmd, GSRES_NORMAL, GS_TIMEOUT);
}

int powerSave (int active, int save) {
    char cmd[GS_CMD_SIZE];

    if (_status != GSSTAT_READY) return -1;

    sprintf(cmd, "AT+WRXACTIVE=%d", active);
    command(cmd, GSRES_NORMAL, GS_TIMEOUT);
    sprintf(cmd, "AT+WRXPS=%d", save);
    return command(cmd, GSRES_NORMAL, GS_TIMEOUT);
}

int standby (int msec) {
    char cmd[GS_CMD_SIZE];

    if (_status != GSSTAT_READY && _status != GSSTAT_WAKEUP) return -1;

    if (_status != GSSTAT_WAKEUP) {
        command("AT+WRXACTIVE=0", GSRES_NORMAL, GS_TIMEOUT);
        command("AT+STORENWCONN", GSRES_NORMAL, GS_TIMEOUT);
    } else {
        command("ATE0", GSRES_NORMAL, GS_TIMEOUT);
        if (_rts) {
            command("AT&K0", GSRES_NORMAL, GS_TIMEOUT);
            command("AT&R1", GSRES_NORMAL, GS_TIMEOUT);
        }
    }
    _status = GSSTAT_STANDBY;
    sprintf(cmd, "AT+PSSTBY=%d,0,0,0", msec); // go standby
    return command(cmd, GSRES_NORMAL, 0);
}

int wakeup () {

    if (_status == GSSTAT_WAKEUP) {
        _status = GSSTAT_READY;
        command("ATE0", GSRES_NORMAL, GS_TIMEOUT);
        if (_rts) {
            command("AT&K0", GSRES_NORMAL, GS_TIMEOUT);
            command("AT&R1", GSRES_NORMAL, GS_TIMEOUT);
        }
#ifdef GS_BULK
        command("AT+BDATA=1", GSRES_NORMAL, GS_TIMEOUT);
#endif
        command("AT+RESTORENWCONN", GSRES_NORMAL, GS_TIMEOUT);
        wait_ms(100);
        return command("AT+WRXACTIVE=1", GSRES_NORMAL, GS_TIMEOUT);
    } else
    if (_status == GSSTAT_DEEPSLEEP) {
        _status = GSSTAT_READY;
        return command("AT", GSRES_NORMAL, GS_TIMEOUT);
    }
    return -1;
}

int deepSleep () {

    if (_status != GSSTAT_READY) return -1;

    _status = GSSTAT_DEEPSLEEP;
    return command("AT+PSDPSLEEP", GSRES_NORMAL, 0); // go deep sleep
}

int wifiConnected () {
    return _connect;
}

enum GSSTATUS getStatus () {
    return _status;
}

int getRssi () {
    if (command("AT+WRSSI=?", GSRES_RSSI, GS_TIMEOUT)) {
        return 0;
    }
    return _rssi;
}

int ntpdate (struct Host *host, int sec) {
    char cmd[GS_CMD_SIZE];

    if (! _connect || _status != GSSTAT_READY) return -1;

    if (host->ip == 0) {
        if (getHostByName2(host)) {
            if (getHostByName2(host)) return -1;
        }
    }

    if (sec) {
    	uint8_t *ip = (uint8_t*)&host->ip;
        sprintf(cmd, "AT+NTIMESYNC=1,%d.%d.%d.%d,%d,1,%d", ip[3], ip[2], ip[1], ip[0],
          GS_TIMEOUT / 1000, sec);
    } else {
    	uint8_t *ip = (uint8_t*)&host->ip;
        sprintf(cmd, "AT+NTIMESYNC=1,%d.%d.%d.%d,%d,0", ip[3], ip[2], ip[1], ip[0],
          GS_TIMEOUT / 1000);
    }
    return command(cmd, GSRES_NORMAL, GS_TIMEOUT);
}

int setTime (char year, char month, char day, char hour, char min, char sec) {
    char cmd[GS_CMD_SIZE];

    if (_status != GSSTAT_READY) return -1;

    sprintf(cmd, "AT+SETTIME=%d/%d/%d,%d:%d:%d", day, month, year + 1900, hour, min, sec);
    return command(cmd, GSRES_NORMAL, GS_TIMEOUT);
}

int getTime (char *year, char *month, char *day, char *hour, char *min, char *sec) {

    if (command("AT+GETTIME=?", GSRES_TIME, GS_TIMEOUT)) {
        return 0;
    }
    *year = _year;
    *month = _month;
    *day = _day;
    *hour = _hour;
    *min = _min;
    *sec = _sec;
    return 1;
}

int gpioOut (int port, int out) {
    char cmd[GS_CMD_SIZE];

    if (_status != GSSTAT_READY) return -1;

    sprintf(cmd, "AT+DGPIO=%d,%d", port, out);
    return command(cmd, GSRES_NORMAL, GS_TIMEOUT);
}

void poll() {
    int i, j;

    if (_gs_enter) {
        // received "\n"
        char buf[GS_CMD_SIZE];

        wait_ms(10);
        _gs_enter --;
        i = 0;
        while (ring_use(_buf_cmd) && i < sizeof(buf)) {
            ring_get(_buf_cmd, &buf[i]);
            if (buf[i] == '\n') {
                if (i == 0) continue;
                break;
            }
            i ++;
        }
        buf[i] = 0;
        DBG("poll: %s\r\n", buf);

        if (strncmp(buf, "CONNECT", 7) == 0 && buf[8] >= '0' && buf[8] <= 'F') {
            i = x2i(buf[8]);
            if (_gs_sock[i].type == GSTYPE_SERVER) {
                int acid, ip1, ip2, ip3, ip4;
                char *tmp = buf + 12;

                acid = x2i(buf[10]);
                newSock(acid, _gs_sock[i].type, _gs_sock[i].protocol, _gs_sock[i].onGsReceive);
                _gs_sock[acid].lcid = i;
                sscanf(tmp, "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4);
                _gs_sock[acid].host.ip = (ip1<<24)|(ip2<<16)|(ip3<<8)|ip4;
                tmp = strstr(tmp, " ") + 1;
                _gs_sock[acid].host.port = atoi(tmp);
            }
        } else
        if (strncmp(buf, "DISCONNECT", 10) == 0) {
            _gs_sock[x2i(buf[11])].connect = 0;
        } else
        if (strncmp(buf, "DISASSOCIATED", 13) == 0 ||
          strncmp(buf, "Disassociated", 13) == 0 ||
          strncmp(buf, "UnExpected", 10) == 0) {
            _connect = 0;
            for (i = 0; i < 16; i ++) {
                _gs_sock[i].connect = 0;
            }
        } else
        if (strncmp(buf, "Out of StandBy-Timer", 20) == 0 ||
          strncmp(buf, "Out of StandBy-Alarm", 20) == 0) {
            if (_status == GSSTAT_STANDBY) {
                _status = GSSTAT_WAKEUP;
            }
        } else
        if (strncmp(buf, "Out of Deep Sleep", 17) == 0 ) {
            if (_status == GSSTAT_DEEPSLEEP) {
                _status = GSSTAT_READY;
            }
        } else
        if (strncmp(buf, "Out of", 6) == 0) {
            if (_status == GSSTAT_STANDBY) {
                _status = GSSTAT_WAKEUP;
            }
            if (_status == GSSTAT_DEEPSLEEP) {
                _status = GSSTAT_READY;
            }
        }
        DBG("status: %d\r\n", _status);
    }

    for (i = 0; i < 16; i ++) {
//        if (_gs_sock[i].connect && _gs_sock[i].received) {
        if (_gs_sock[i].received && ring_use(_gs_sock[i].data)) {
            // recv interrupt
            _gs_sock[i].received = 0;
        	if (_gs_sock[i].onGsReceive != NULL) {
                for (j = 0; j < 1500 / GS_DATA_SIZE + 1; j ++) {
                    if (! ring_use(_gs_sock[i].data)) break;
                    _gs_sock[_cid].onGsReceive(_cid, ring_use(_gs_sock[i].data));
                }
            }
        }
    }
}

void newSock (int cid, enum GSTYPE type, enum GSPROTOCOL pro, onGsReceiveFunc ponGsReceive) {
    _gs_sock[cid].type = type;
    _gs_sock[cid].protocol = pro;
    _gs_sock[cid].connect = 1;
    if (_gs_sock[cid].data == NULL) {
        _gs_sock[cid].data = ring_new(GS_DATA_SIZE);
    } else {
        ring_clear(_gs_sock[cid].data);
    }
    _gs_sock[cid].lcid = 0;
    _gs_sock[cid].received = 0;
    _gs_sock[cid].onGsReceive = ponGsReceive;
}

int open (struct Host *host, enum GSPROTOCOL pro, onGsReceiveFunc ponGsReceive) {
    char cmd[GS_CMD_SIZE];

    if (! _connect || _status != GSSTAT_READY) return -1;
    if (host->ip == 0 || host->port == 0) {
        return -1;
    }

    if (pro == GSPROT_UDP) {
    	uint8_t *ip = (uint8_t*)&host->ip;
        sprintf(cmd, "AT+NCUDP=%d.%d.%d.%d,%d", ip[3], ip[2], ip[1], ip[0], host->port);
    } else {
    	uint8_t *ip = (uint8_t*)&host->ip;
        sprintf(cmd, "AT+NCTCP=%d.%d.%d.%d,%d", ip[3], ip[2], ip[1], ip[0], host->port);
    }
    if (command(cmd, GSRES_CONNECT, GS_TIMEOUT)) return -1;

    newSock(_cid, GSTYPE_CLIENT, pro, ponGsReceive);
    return _cid;
}

int listen (int port, enum GSPROTOCOL pro, onGsReceiveFunc ponGsReceive) {
    char cmd[GS_CMD_SIZE];

    if (! _connect || _status != GSSTAT_READY) return -1;
    if (port == 0) {
        return -1;
    }

    if (pro == GSPROT_UDP) {
        sprintf(cmd, "AT+NSUDP=%d", port);
    } else {
        sprintf(cmd, "AT+NSTCP=%d", port);
    }
    if (command(cmd, GSRES_CONNECT, GS_TIMEOUT)) return -1;

    newSock(_cid, GSTYPE_SERVER, pro, ponGsReceive);
    return _cid;
}

int close (int cid) {
    char cmd[GS_CMD_SIZE];

    if (! _gs_sock[cid].connect) return -1;

    _gs_sock[cid].connect = 0;
    _gs_sock[cid].data = NULL;
    sprintf(cmd, "AT+NCLOSE=%X", cid);
    return command(cmd, GSRES_NORMAL, GS_TIMEOUT);
}

int send (int cid, char *buf, int len) {
    int i;
    char s[40];

    if (! _gs_sock[cid].connect) return -1;

    if ((_gs_sock[cid].protocol == GSPROT_TCP) ||
      (_gs_sock[cid].protocol == GSPROT_UDP && _gs_sock[cid].type == GSTYPE_CLIENT)) {
        // TCP Client, TCP Server, UDP Client
        _gs_ok = 0;
        _gs_failure = 0;
#ifdef GS_BULK
        sprintf(s, "\x1bZ%X%04d", cid, len);
        _gs_printf(s);
        for (i = 0; i < len; i ++) {
            _gs_putc(buf[i]);
            DBG("%c", buf[i]);
        }
#else
        sprintf(s, "\x1bS%X", cid);
        _gs_printf(s);
        for (i = 0; i < len; i ++) {
            if (buf[i] >= 0x20 && buf[i] < 0x7f) {
                _gs_putc(buf[i]);
                DBG("%c", buf[i]);
            }
        }
        _gs_putc(0x1b);
        _gs_putc('E');
#endif
    } else {
        return -1;
    }
    Timer_start();
    while (!_gs_ok && !_gs_failure && Timer_read_ms() < GS_TIMEOUT);
    return _gs_ok == 1 ? 0 : -1;
}

int sendto (int cid, char *buf, int len, struct Host *host) {
    int i;
    char s[40];
    uint8_t *ip = (uint8_t *)&host->ip;

    if (! _gs_sock[cid].connect) return -1;

    if ((_gs_sock[cid].protocol == GSPROT_UDP && _gs_sock[cid].type == GSTYPE_SERVER)) {
        // UDP Server
        _gs_ok = 0;
        _gs_failure = 0;
#ifdef GS_BULK
        sprintf(s, "\x1bY%X", cid);
        _gs_printf(s);
        sprintf(s, "%d.%d.%d.%d:%d:", ip[3], ip[2], ip[1], ip[0], host->port);
        _gs_printf(s);
        sprintf(s, "%04d", len);
        _gs_printf(s);
        for (i = 0; i < len; i ++) {
            _gs_putc(buf[i]);
            DBG("%c", buf[i]);
        }
#else
        sprintf(s, "\x1bU%X", cid);
        _gs_printf(s);
        sprintf(s, "%d.%d.%d.%d:%d:", host.getIp()[0], host.getIp()[1], host.getIp()[2], host.getIp()[3], host.getPort());
        _gs_printf(s);
        for (i = 0; i < len; i ++) {
            if (buf[i] >= 0x20 && buf[i] < 0x7f) {
                _gs_putc(buf[i]);
                DBG("%c", buf[i]);
            }
        }
        _gs_putc(0x1b);
        _gs_putc('E');
#endif
    } else {
        return -1;
    }
    Timer_start();
    while (!_gs_ok && !_gs_failure && Timer_read_ms() < GS_TIMEOUT);
    return _gs_ok == 1 ? 0 : -1;
}

int recv (int cid, char *buf, int len) {
    int r;

    if (_gs_sock[cid].data == NULL) return 0;

    Timer_start();
    while (ring_use(_gs_sock[cid].data) == 0) {
        if (Timer_read_ms() > GS_TIMEOUT) return 0;
    }
    Timer_stop();

    r = ring_gets(_gs_sock[cid].data, buf, len);
    return r;
}

int recvfrom (int cid, char *buf, int len, struct Host *host) {
    int r;

    if (_gs_sock[cid].data == NULL) return 0;

    Timer_start();
    while (ring_use(_gs_sock[cid].data) == 0) {
        if (Timer_read_ms() > GS_TIMEOUT) return 0;
    }
    Timer_stop();

    r = ring_gets(_gs_sock[cid].data, buf, len);
    memcpy(host, &_from, sizeof(host));
    return r;
}

int isConnected (int cid) {
    return _gs_sock[cid].connect;
}

int httpGet (struct Host *host, char *uri, int ssl, onGsReceiveFunc ponGsReceive) {
    char cmd[GS_CMD_SIZE];
    uint8_t *ip = (uint8_t *)&host->ip;

    if (! _connect || _status != GSSTAT_READY) return -1;

    if (host->ip == 0) {
        if (getHostByName2(host)) {
            if (getHostByName2(host)) return -1;
        }
    }
    if (host->port == 0) {
        if (ssl) {
            host->port = 443;
        } else {
            host->port = 80;
        }
    }

    command("AT+HTTPCONF=3,close", GSRES_NORMAL, GS_TIMEOUT); // Connection:
    sprintf(cmd, "AT+HTTPCONF=11,%s", host->name);  // Host:
    command(cmd, GSRES_NORMAL, GS_TIMEOUT);
    sprintf(cmd, "AT+HTTPOPEN=%d.%d.%d.%d,%d", ip[3], ip[2], ip[1], ip[0], host->port);
    if (ssl) {
        strcat(cmd, ",1");
    }
    if (command(cmd, GSRES_HTTP, GS_TIMEOUT)) return -1;

    newSock(_cid, GSTYPE_CLIENT, GSPROT_HTTPGET, ponGsReceive);

    sprintf(cmd, "AT+HTTPSEND=%d,1,%d,%s", _cid, GS_TIMEOUT / 1000, uri);  // Get:
    command(cmd, GSRES_NORMAL, GS_TIMEOUT);

    return _cid;
}

int certAdd (char *name, char *cert, int len) {
    int i;
    char cmd[GS_CMD_SIZE];

    if (! _connect || _status != GSSTAT_READY) return -1;

    sprintf(cmd, "AT+TCERTADD=%s,1,%d,1", name, len);  // Hex, ram
    command(cmd, GSRES_NORMAL, GS_TIMEOUT);

    _gs_putc(0x1b);
    _gs_putc('W');
    for (i = 0; i < len; i ++) {
        _gs_putc(cert[i]);
    }
    return cmdResponse(GSRES_NORMAL, GS_TIMEOUT);
}

int x2i (char c) {
    if (c >= '0' && c <= '9') {
        return c - '0';
    } else
    if (c >= 'A' && c <= 'F') {
        return c - 'A' + 10;
    } else
    if (c >= 'a' && c <= 'f') {
        return c - 'a' + 10;
    }
    return 0;
}

char i2x (int i) {
    if (i >= 0 && i <= 9) {
        return i + '0';
    } else
    if (i >= 10 && i <= 15) {
        return i - 10 + 'A';
    }
    return 0;
}


struct RingBuffer *ring_new (int p_size) {
    struct RingBuffer *ring;
    ring = (struct RingBuffer *)malloc(sizeof(struct RingBuffer));
    ring->size = p_size + 1;
    ring->buf = (char*)malloc(ring->size);
    ring->addr_w = 0;
    ring->addr_r = 0;
    return ring;
}

int ring_put (struct RingBuffer *ring, char dat) {
    int next;

    next = (ring->addr_w + 1) % ring->size;
    if (next == ring->addr_r) {
        return -1;
    }
    ring->buf[ring->addr_w] = dat;
    ring->addr_w = next;
    return -1;
}

int ring_get (struct RingBuffer *ring, char *dat) {
    if (ring->addr_r == ring->addr_w) {
        return -1;
    }
    *dat = ring->buf[ring->addr_r];
    ring->addr_r = (ring->addr_r + 1) % ring->size;
    return 0;
}

int ring_gets (struct RingBuffer *ring, char *dat, int len) {
    int i;

    for (i = 0; i < len; i ++) {
        if (ring->addr_r == ring->addr_w) {
            break;
        }
        dat[i] = ring->buf[ring->addr_r];
        ring->addr_r = (ring->addr_r + 1) % ring->size;
    }
    return i;
}

int ring_available (struct RingBuffer *ring) {
    if (ring->addr_w < ring->addr_r) {
        return ring->addr_r - ring->addr_w - 1;
    } else {
        return (ring->size - ring->addr_w) + ring->addr_r - 1;
    }
}

int ring_use (struct RingBuffer *ring) {
    return ring->size - ring_available(ring) - 1;
}

void ring_clear (struct RingBuffer *ring) {
    ring->addr_w = 0;
    ring->addr_r = 0;
}
