/**
 * Gainspan wi-fi module library for LPC1114 (LPCXpresso)
 * code from http://mbed.org/users/gsfan/code/GSwifi/
 * Modified for Arduino, 2012 Galileo-7, S.
 */

/*
 * Gainspan wi-fi module library for mbed
 * Copyright (c) 2012 gsfan
 * Released under the MIT License: http://mbed.org/license/mit
 */

/** @file
 * @brief Gainspan wi-fi module library for mbed
 * GS1011MIC, GS1011MIP, GainSpan WiFi Breakout, etc.
 * module configuration: ATB=115200
 */

#ifndef _H_
#define _H_

//#define GS_BAUD 115200
#define GS_BAUD 9600
#define GS_UART_DIRECT

#define GS_BULK

#define GS_TIMEOUT 10000 // ms
#define GS_TIMEOUT2 30000 // ms
#define GS_CMD_SIZE 100
#define GS_DATA_SIZE 256

/**
 * Wi-Fi security
 */
enum GSSECURITY {
    GSSEC_AUTO = 0,
    GSSEC_NONE = 0,
    GSSEC_OPEN = 1,
    GSSEC_WEP = 2,
    GSSEC_WPA_PSK = 4,
    GSSEC_WPA2_PSK = 8,
    GSSEC_WPA_ENT = 16,
    GSSEC_WPA2_ENT = 32,
    GSSEC_WPS_BUTTON = 64,
};

/**
 * TCP/IP protocol
 */
enum GSPROTOCOL {
    GSPROT_UDP = 0,
    GSPROT_TCP = 1,
    GSPROT_HTTPGET,
    GSPROT_HTTPPOST,
};

/**
 * Client/Server
 */
enum GSTYPE {
    GSTYPE_CLIENT = 0,
    GSTYPE_SERVER = 1,
};

enum GSRESPONCE {
    GSRES_NORMAL,
    GSRES_CONNECT,
    GSRES_WPS,
    GSRES_MACADDRESS,
    GSRES_DHCP,
    GSRES_DNSLOOKUP,
    GSRES_HTTP,
    GSRES_RSSI,
    GSRES_TIME,
};

enum GSMODE {
    GSMODE_COMMAND,
    GSMODE_DATA_RX,
    GSMODE_DATA_RXUDP,
    GSMODE_DATA_RX_BULK,
    GSMODE_DATA_RXUDP_BULK,
//    GSMODE_DATA_RXHTTP,
};

enum GSSTATUS {
    GSSTAT_READY,
    GSSTAT_STANDBY,
    GSSTAT_WAKEUP,
    GSSTAT_DEEPSLEEP,
};

/**
 * data receive callback function
 */
typedef void (*onGsReceiveFunc)(int cid, int len);

struct RingBuffer {
    char *buf;
    int size;
    int addr_w, addr_r;
};
typedef uint32_t IpAddr;
struct Host {
	char *name;
	IpAddr ip;
	int port;
};

struct GS_Socket {
	enum GSTYPE type;
	enum GSPROTOCOL protocol;
    int connect;
    struct Host host;
    struct RingBuffer *data;
    int lcid;
    int received;
    onGsReceiveFunc onGsReceive;
};

/**
 * GSwifi class
 */
    /**
     * default constructor
     */
    void GSwifi (int baud, int flow);
    /**
     * Default constructor (with hardware fllow controll)
     */
    int command (const char *cmd, enum GSRESPONCE res, int timeout);
    /**
     * recv responce
     */
    int cmdResponse (enum GSRESPONCE res, int ms);
    /**
     * associate infrastructure
     * @param sec GSSEC_OPEN, GSSEC_WEP, GSSEC_WPA_PSK, GSSEC_WPA2_PSK, GSSEC_WPS_BUTTON
     * @param ssid SSID
     * @param pass pass phrase
     * @param dhcp 0:static ip, 1:dhcp
     * @retval 0 success
     * @retval -1 failure
     */
    int connect (enum GSSECURITY sec, char *ssid, char *pass, int dhcp);
    /**
     * adhock
     * @param sec GSSEC_OPEN or GSSEC_WEP
     * @param ssid SSID
     * @param pass 10 or 26 hexadecimal digits
     * @param ipaddr my ip address
     * @param netmask subnet mask
     * @retval 0 success
     * @retval -1 failure
     */
    int adhock (enum GSSECURITY sec, char *ssid, char *pass, IpAddr ipaddr, IpAddr netmask);
    /**
     * limited AP
     * @param sec GSSEC_OPEN or GSSEC_WEP
     * @param ssid SSID
     * @param pass 10 or 26 hexadecimal digits
     * @param ipaddr my ip address (dhcp start address)
     * @param netmask subnet mask
     * @retval 0 success
     * @retval -1 failure
     * firmware: s2w-secureweb, s2w-web, s2w-wpsweb
     */
    int limitedap (enum GSSECURITY sec, char *ssid, char *pass, IpAddr ipaddr, IpAddr netmask);
    /**
     * unassociate
     */
    int disconnect ();

    /**
     * use static ip address
     */
    int setAddress (IpAddr ipaddr, IpAddr netmask, IpAddr gateway, IpAddr nameserver);
    /**
     * get ip address
     */
    int getAddress (IpAddr *ipaddr, IpAddr *netmask, IpAddr *gateway, IpAddr *nameserver);
    /**
     * resolv hostname
     * @param name hostname
     * @param addr resolved ip address
     */
    int getHostByName (const char* name, IpAddr *addr);
    /**
     * resolv hostname
     * @param host.name hostname
     * @param host.ipaddr resolved ip address
     */
    int getHostByName2 (struct Host *host);
    /**
     * RF power
     * @param power 0(high)-7(low)
     */
    int setRFPower (int power);
    /**
     * power save mode
     * @param active rx radio 0:switched off, 1:always on
     * @param save power save 0:disable, 1:enable
     */
    int powerSave (int active, int save);
    /**
     * standby mode
     * @param msec wakeup after
     * wakeup after msec or alarm1/2
     * core off, save to RTC ram
     */
    int standby (int msec);
    /**
     * restore standby
     */
    int wakeup ();
    /**
     * deep sleep mode
     */
    int deepSleep ();
    /**
     * wifi connected
     */
    int wifiConnected ();
    /**
     * status
     * @return GSSTATUS
     */
    enum GSSTATUS getStatus ();
    /**
     * RSSI
     * @return RSSI (dBm)
     */
    int getRssi ();
    /**
     * set NTP server
     * @param host SNTP server
     * @param sec time sync interval, 0:one time
     */
    int ntpdate (struct Host *host, int sec);
    /**
     * set system time
     * @param time date time (UTC)
     */
    int setTime (char year, char month, char day, char hour, char min, char sec);
    /**
     * get RTC time
     * @return date time (UTC)
     */
    int getTime (char *year, char *month, char *day, char *hour, char *min, char *sec);
    /**
     * GPIO output
     * @param port 10,11,30,31
     * @param out 0:set(high), 1:reset(low)
     */
    int gpioOut (int port, int out);

    /**
     * main polling
     */
    void poll();

    /**
     * tcp/udp client
     * @return CID, -1:failure
     */
    int open (struct Host *host, enum GSPROTOCOL pro, onGsReceiveFunc ponGsReceive);
    /**
     * tcp/udp server
     * @return CID, -1:failure
     */
    int listen (int port, enum GSPROTOCOL pro, onGsReceiveFunc ponGsReceive);
    /**
     * close client/server
     */
    int close (int cid);
    /**
     * send data tcp(s/c), udp(c)
     */
    int send (int cid, char *buf, int len);
    /**
     * send data udp(s)
     */
    int sendto (int cid, char *buf, int len, struct Host *host);
    /**
     * recv data tcp(s/c), udp(c)
     * @return length
     */
    int recv (int cid, char *buf, int len);
    /**
     * recv data udp(s)
     * @return length
     */
    int recvfrom (int cid, char *buf, int len, struct Host *host);
    /**
     * tcp/udp connected
     */
    int isConnected (int cid);

    /**
     * http request
     * @return CID, -1:failure
     * If you use ssl, please set system time.
     */
    int httpGet (struct Host *host, char *uri, int ssl, onGsReceiveFunc ponGsReceive);
    /**
     * certificate
     */
    int certAdd (char *name, char *cert, int len);


    int x2i (char c);
    char i2x (int i);
    void newSock (int cid, enum GSTYPE type, enum GSPROTOCOL pro, onGsReceiveFunc ponGsReceive);

    void init_uart (int baud, int flow);
    void _gs_printf (const char *buf);
    struct RingBuffer *ring_new (int p_size);
    int ring_put (struct RingBuffer *ring, char dat);
    int ring_get (struct RingBuffer *ring, char *dat);
    int ring_gets (struct RingBuffer *ring, char *dat, int len);
    void ring_clear (struct RingBuffer *ring);
    int ring_available (struct RingBuffer *ring);
    int ring_use (struct RingBuffer *ring);


#endif
