#include "dbg.h"
#include "LPC11xx.h"
#include <cr_section_macros.h>
#include <NXP/crp.h>

#include <stdlib.h>
#include <string.h>
#include "GSwifi.h"
#include "7seg.h"
#include "temp.h"

#define HTTP_HOST "www.kepco.co.jp"
#define HTTP_URI "/yamasou/juyo1_kansai.csv"
#define SSID "SSID"
#define PASS "password"
#define NTP_HOST "ntp.jst.mfeed.ad.jp"

#define GS_RESET_H LPC_GPIO1->DATA |= (1<<1)
#define GS_RESET_L LPC_GPIO1->DATA &= ~(1<<1)
#define GS_WAKEUP_H LPC_GPIO1->DATA |= (1<<0)
#define GS_WAKEUP_L LPC_GPIO1->DATA &= ~(1<<0)

__CRP const unsigned int CRP_WORD = CRP_NO_CRP ;

volatile uint32_t time_millis = 0, time_micros = 0;
volatile int denki_flg = 0;
volatile int denki_capacity = 0;
volatile int denki_usage = 0;
volatile int year, month, day, hour, minute;

uint32_t millis () {
	return time_millis;
}

void wait_ms(int ms) {
	int n = millis() + ms;
	while (millis() <= n);
	return;
}

// 500Hz (2ms)
void SysTick_Handler(void) {
	time_millis += 2;
	time_micros += 500;
	if (time_micros >= 1000) {
		time_millis ++;
		time_micros = 0;
	}
	leddisp();
}

void error () {
	volatile int i;
	writeled(0, 4, 0x0f, 0x0f);
	for (;;) {
		LPC_GPIO1->DATA |= (1<<5); // led on
		for (i = 0; i < 100000; i ++);
		LPC_GPIO1->DATA &= ~(1<<5); //led off
		for (i = 0; i < 100000; i ++);
	}
}

void callback_http (int cid, int len) {
    static int n = 0;
//    static char buf[1024], data[80];
    static char buf[GS_DATA_SIZE], data[80];
    int i;
    int last = len;

  while (last) {
    len = recv(cid, buf, last < sizeof(buf) ? last : sizeof(buf));
    last = last - len;

    for (i = 0; i < len; i ++) {
        if (buf[i] == '\r') continue;

        if (denki_flg <= 1) {
            // header
            if (strncmp(buf, "200 OK", 6) == 0) {
                i += 7;
                denki_flg = 2;
            }
            continue;
        }

        // body
        if (buf[i] == '\n') {
            data[n] = 0;
            switch (denki_flg) {
            case 2:
                // update
                break;
            case 4:
                // capacity
                if (data[0] >= '0' && data[0] <= '9') {
                    denki_capacity = atoi(data);
                }
                break;
            case 7:
                // yosou1
                break;
            case 10:
                // yosou2
                break;
            default:
                // text
                if (data[0] == 'D' && data[1] == 'A') {
                    denki_flg = denki_flg < 100 ? 100 : 200;
                }
                break;
            }
            if (denki_flg > 200 && data[n - 1] >= '0' && data[n - 1] <= '9') {
                sscanf(data, "%d/%d/%d,%d:%d,%d", &year, &month, &day, &hour, &minute, &denki_usage);
            }
            n = 0;
            denki_flg ++;
        } else {
            // data
            if (n < sizeof(data) - 1) {
                data[n] = buf[i];
                n ++;
            }
        }

    }

  }
}

void Sleep () {
    __WFI();
}

int main(void) {
    int flg = 1, r;
    float denki_percentage = 0;
    struct Host host, ntp;
    char name_ntp[] = NTP_HOST;
    char name_http[] = HTTP_HOST;
    int count = 0;
    char _year, _month, _day, _hour, _min, _sec;

	NVIC_EnableIRQ(SysTick_IRQn);
	SysTick_Config(SystemCoreClock / 400); // 400Hz

	initled();
	init_temp();
	LPC_GPIO1->DIR |= (1<<1)|(1<<0);
   	LPC_IOCON->R_PIO1_0 |= 0x01;
   	LPC_IOCON->R_PIO1_1 |= 0x01;
	GS_RESET_L;
    GS_WAKEUP_L;
    wait_ms(100);
    GS_WAKEUP_H;
    GS_RESET_H;
    wait_ms(500);

	GSwifi(115200, 1);
	if (connect(GSSEC_WPA2_PSK, SSID, PASS, 1)) {
		error();
	}

    ntp.name = name_ntp;
    ntp.ip = 0;
    ntp.port = 0;
    ntpdate(&ntp, 0);
    wait_ms(15000);
    ntpdate(&ntp, 3 * 60 * 60);

    host.name = name_http;
    host.ip = 0;
    host.port = 0;

    while(1) {
        poll();

        if (flg || getStatus() == GSSTAT_WAKEUP) {
            if (getStatus() == GSSTAT_WAKEUP) {
                wakeup();
            }

            r = httpGet(&host, HTTP_URI, 0, &callback_http);
            if (r >= 0) {
                denki_percentage = (float)denki_usage / (float)denki_capacity * 100.0;
            }
            flg = 0;

            wait_ms(5000);
            getTime(&_year, &_month, &_day, &_hour, &_min, &_sec);
            _hour += 9;
            if (_hour >= 24) {
            	_hour -= 24;
            	_day ++;
            }

            standby(180000);
        }

        if (count == 0) { // 0sec
            // time
            writeled(_hour * 100 + _min, 2, 0x0f, 0x0f);
        } else
        if (count > 0 && count < 4000) {
            // time blink
            if (count % 400 == 0) { // 0sec
                writeled(_hour * 100 + _min, 2, 0x0f, 0x0f);
            } else
            if (count % 400 == 200) { // 0.1sec
                writeled(_hour * 100 + _min, 0, 0x0f, 0x0f);
            }
        } else
        if (count == 4000) { // 10sec
            // denki yohou
            writeled(denki_percentage * 100, 2, 0x0e, 0x02);
        } else
        if (count == 6000) { // 15sec
            // temp.
            writeled(get_temp() * 10, 3, 0x07, 0);
        }

        if (count % 400 == 0) { // 0sec
        } else
        if (count % 400 == 40) { // 0.1sec
            _sec ++;
            if (_sec >= 60) {
            	_sec = 0;
            	_min ++;
            	if (_min >= 60) {
            		_min = 0;
            		_hour ++;
            		if (_hour >= 24) {
            			_hour = 0;
            			_day ++;
            		}
            	}
            }
        }

        count ++;
        if (count >= 8000) count = 0; // 20sec

        Sleep();
    }
}
