#include "LPC11xx.h"
#include "temp.h"
#include <math.h>

#define VREF 3.3
#define B 3435 // thermistor B
#define T0 25
#define R0 10000 // R0 ohm
#define RU 10000 // pullup ohm

void init_temp () {
	LPC_IOCON->R_PIO0_11 &= ~((1<<7)|(3<<3));
	LPC_IOCON->R_PIO0_11 |= (2<<0);
	LPC_SYSCON->PDRUNCFG &= ~(1<<4); // ADC_PD
	LPC_SYSCON->SYSAHBCLKCTRL |= (1<<13);
	LPC_ADC->CR = (1<<24)|(10<<8)|(1<<0);
}

int get_adc (int ch) {
	uint32_t ad;
	LPC_ADC->CR = (1<<24)|(10<<8)|(1<<ch);
	for (;;) {
		ad = LPC_ADC->DR[ch];
		if (ad & (1<<31)) break;
	}
	LPC_ADC->CR &= ~(1<<24);
	return (ad >> 6) & 0x3ff;
}

float get_temp () {
    float v, r, t;

    v = (float)get_adc(0) / 1024 * VREF;
    r = v / ((VREF - v) / RU);
    t = (1.0 / ((1.0 / (T0 + 273.15)) + (log(r / R0) / B))) - 273.15;
    if (t < -50) t = 0;
    return t;
}
